#!/bin/bash

# v0.1 - 2021.04.13 - derived from old configure_eth_rxq_irq_rss_fdir.sh, 
#
#                   only about configure FDir.
#
#
# v0.2 - 2021.04.21 - rewrite, to adapt to both Bash and embedded busybox ash
#
#
# v0.3 - 2021.06.05 - add workaround logic, called "sw_fdir_post", to handle this problem:
#
#							on Intel i7/i9 CPU, frequent IPI from non-CPU-0 would cause
#							kernel timer_list trigger inaccuracy.
#
#							see bug #24865
#
#
# v0.4 - 2021.06.28 - handle FDir on virtual interface
#
#
#		when set on FDir on a virtual interface, then enable FDir on all of its lower physical interfaces
#
#
#		when set off FDir on a virtual interface, need to check if it has sibiling, or lower parent, or upper child
#		in wan-tbl. If yes, then don't set off its lower physical interfaces.
#
#		when set off FDir on a physical interface, need to check if it has upper child in wan-tbl.
#
#
#		An example interface relationship is like:
#
#				( eth0, eth1 ) -> bond0 -> ( bond0.100, bond0.200 )
#
#

#
# This script is to:
#
#       configure net_device's FDir
#
#       
#       either HW FDir ( i40e only ), or SW FDir ( other NIC without HW FDir )
#
#
# Usage:
#
#       this_script  all  [on|off] [hw|sw]
#
#           configure all net_device to be on | off.
#
#           hw | sw
#
#               when set to 'on' and 'hw', then if a net_device support HW FDir ( only for i40e)
#               then try to configure its HW FDir, otherwise, configure its SW FDir.
#
#               when set to 'on' and 'sw', then configure their SW FDir even it support HW FDir.
#
#               when not specific 'hw' or 'sw', then, for i40e-only, configure HW FDir
#               for other, configure SW FDir
#
#
#               when set to 'off', hw or sw is mearningless, turn off both HW FDir and SW FDir.
#
#
#       this_script  ethX [on|off] [hw|sw]
#
#           configure the specified net_device. 
#
#
# The 1st usage should be called in system init time, scanning all interface and configuring them.
#
#
# The 2nd usage should be called when a interface become WAN.
#


gLogFile="/var/log/messages"
gCurrentTime=`date`
gLogPrefix="${gCurrentTime} <LWLog:Info:LW_FDIR>"

log()
{
	echo "${gLogPrefix}: $@" >> ${gLogFile}
}


###########################################
#
# MUST accept 3 argumments.
#
###########################################

if [ $# != 3 ]; then
    log "Usage: $0 [all|ethX] [on|off] [hw|sw]"
    exit 1
fi


#
# check if running on busybox ash, probably arm/mips CPE
#

is_running_on_busybox=0
if [ -e /bin/busybox ]; then
	is_running_on_busybox=1
fi


#
# check if configuring a single interface, instead of all interface
#

is_configuring_a_single_interface=0
if [ "$1" != "all" ]; then
	is_configuring_a_single_interface=1
fi


#
# global variables as control flag, used when set off FDir
#

gSetOffCheckSiblingInWanTbl=0
gSetOffCheckLowerInWanTbl=0
gSetOffCheckUpperInWanTbl=0

if [ ${is_configuring_a_single_interface} -eq 1 ]; then

	gSetOffCheckSiblingInWanTbl=1
	gSetOffCheckLowerInWanTbl=1
	gSetOffCheckUpperInWanTbl=1

fi


#
# When set off FDir to a physical interface from upper virtual interface, 
#
# Use this flag to avoid check gSetOffCheckUpperInWanTbl recursively.
#

gSetOffPhyFromVirtInf=0


##########################################
#
# get cpu_num ( system CPU number )
#
##########################################

cpu_num=`cat /proc/cpuinfo | grep "processor" | wc -l` > /dev/null
if [ ${cpu_num} -eq 1 ]; then
    log "system cpu_num: 1, no need to configure fdir"
    exit 0
fi


log "system cpu_num: ${cpu_num}"



###########################################
#
# variable declare.
#
###########################################

#
# drivers supports HW FDir or SW FDir
#

phy_drivers="i40e ixgbe igb e1000e fsl_dpa mtk_soc_eth mvneta dsa r8168 st_gmac"



###########################################
#
# sw_fdir_post
#
#
#   This is a workaround for bug #24865 ( frequent IPI would cause kernel timer triggering abnormal ), 
#	which is observed on Intel x86 CPU ( multi CPU core > 2 ):
#	
#		If sending frequent network RPS IPI from physical-CPU-non-0 to other physical-CPU-non-0, 
# 		then would cause kernel timer trigger delay on both CPUs.
#
#
#		Note, here we emphasize "phyiscal-CPU". because if with HT (Hyper-Threading) on, 
#	a physical-CPU behave as more logic CPU. For example, a physical-CPU-1 correspond to:
#	logical-CPU-1 and logical-CPU-9 on a i9_9900K system.
#
#
#		This is because: 
#
#			physical-CPU-0 is the tick master updating jiffies.
#
#			While other physical-CPU-0 execute their timer based on jiffies.
#
#			network RPS IPI logic cause some irq disable and lock contention between 2 CPU involved.
#
#			If RPS IPI between CPU-a and CPU-b, then these 2 CPU ( in fact, all related logical CPUs )
#			tends to be slower than CPU-0 updating jiffies. Then the timers on these 2 CPUs
#			happen delay.
#
#			But if RPS IPI between CPU-0 and other CPUs, then CPU-0 itself is slower like other 
#			CPU, jiffies updating is slower, so no timer delay.
#
#
#
#	So the workaround is: 
#
#
#		Any logical-CPU belong to physical-CPU-0 is considered good CPU.
#
#
#		when turn on SW FDir on Intel NIC and CPU, we force all of its RXQ bound to these good CPUs.
#
#
###########################################

is_running_on_intel_cpu=0

tmpvar=`cat /proc/cpuinfo | grep "Intel" | wc -l` > /dev/null

if [ $tmpvar -ge 1 ]; then
	is_running_on_intel_cpu=1
fi


#
# check if Intel HT(Hyper-Threading) on,
#
# and get "Thread(s) per core:"
#

is_running_on_intel_cpu_ht_on=0

lscpu_thread_per_core=1
lscpu_thread_per_core=`lscpu | grep "Thread(s) per core" | sed 's/Thread(s) per core://g'` > /dev/null
lscpu_thread_per_core=$(echo "${lscpu_thread_per_core}" | sed 's/[[:space:]]//g')

if [ ${lscpu_thread_per_core} -gt 1 ]; then
	is_running_on_intel_cpu_ht_on=1
fi


#
# only Intel CPU ( multi CPU core > 2 ) need to check and do such post work.
#

need_sw_fdir_post_on_this_system=0

if [ ${is_running_on_intel_cpu} -eq 1 ]; then

	if [ ${cpu_num} -gt 2 ]; then

		is_running_on_intel_cpu_powerful=`cat /proc/cpuinfo | grep -E "(i7-)|(i9-)" | wc -l` > /dev/null

		if [ ${is_running_on_intel_cpu_powerful} -ge 1 ]; then

			log "Intel CPU(i7, i9), may be affected by #24865, need to do SW FDir post work for igb/ixgbe"

			need_sw_fdir_post_on_this_system=1

		else

			log "Your CPU type not affected by #24865, no need to do SW FDir post work"

			need_sw_fdir_post_on_this_system=0

		fi

	fi
fi

log "is_running_on_intel_cpu=${is_running_on_intel_cpu}"
log "is_running_on_intel_cpu_ht_on=${is_running_on_intel_cpu_ht_on}"
log "lscpu_thread_per_core=${lscpu_thread_per_core}"
log "need_sw_fdir_post_on_this_system=${need_sw_fdir_post_on_this_system}"


#
# these variables are ONLY used in need_sw_fdir_post_on_this_system==1
#

sw_fdir_post_nic_multi_rxq_num=1

sw_fdir_post_nic_rxq_cpu_id_step=${cpu_num}


if [ ${need_sw_fdir_post_on_this_system} -eq 1 ]; then

	#
	# If Intel HT is on, then there is 2 logical-CPU corresponding to 1 physical-CPU
	# For example:
	#			On a 16 core system, logical-CPU-0 and logical-CPU-8 
	#			corresponds to physical-CPU-0.
	#			We can use these 2 logical CPU.
	#

	if [ ${is_running_on_intel_cpu_ht_on} -eq 1 ]; then

		sw_fdir_post_nic_multi_rxq_num=2

		sw_fdir_post_nic_rxq_cpu_id_step=`expr ${cpu_num} / 2`

	else

		sw_fdir_post_nic_multi_rxq_num=1

		sw_fdir_post_nic_rxq_cpu_id_step=${cpu_num}	

	fi

fi


log "sw_fdir_post_nic_multi_rxq_num=${sw_fdir_post_nic_multi_rxq_num}"
log "sw_fdir_post_nic_rxq_cpu_id_step=${sw_fdir_post_nic_rxq_cpu_id_step}"


#
# physical NIC drivers can and should do sw_fdir_post
#
#
#	Note, normally i40e is using its HW Fdir, and agent only configure its HW Fdir.
#
#	But we also implement SW Fdir for i40e for test purpose. So here also include i40e
#
#	But i40e SW Fdir would not actually use in real scenario.
#

need_sw_fdir_post_on_this_system_phy_drivers_intel_nic_multi_rxq="i40e ixgbe igb"
need_sw_fdir_post_on_this_system_phy_drivers_intel_nic_single_rxq="e1000e"


##########################################
#
# functions
#
##########################################


phy_configure_fdir_on_sw_post_intel_nic_multi_rxq()
{

    ifname=$1
    driver_name=$2

	log "phy_configure_fdir_on_sw_post_intel_nic_multi_rxq(): ${ifname}/${driver_name}"

	log "will forcely set its RXQ to be 1 or 2, depending on CPU HT off / on"
	log "and bind RXQ to CPU-0 or/and sibiling CPU if CPU HT on"


	#
	# set interface RXQ to be 1 or 2, depending on HT off / on
	#
	# #31463 the following setting command would cause interface up/down, because the combined value is not same
	# in configure_eth_fdir.sh and configure_eth_rxq_irq_rss.sh, so we comment out the following command.

	#log "ethtool --set-channels ${ifname} combined ${sw_fdir_post_nic_multi_rxq_num}"

	#ethtool --set-channels ${ifname} combined ${sw_fdir_post_nic_multi_rxq_num} > /dev/null 2>&1


	#
	# bind RXQ to CPU-0 and CPU-4 or/CPU-8 or/CPU-16 if HT on
	#
	#
	#	intel_nic_multi_rxq's IRQ naming pattern is:
	#
	#		eth0-TxRx-N
	#

	irq_ids_num=`cat /proc/interrupts | grep ${ifname}-TxRx | wc -l` > /dev/null

	irq_ids=`cat /proc/interrupts | grep ${ifname}-TxRx | awk '{print $1}' | sed 's/://g'` > /dev/null

	#
	# this is also to check if previous "ethtool --set-channels" success
	#

	#if [ ${irq_ids_num} -ne ${sw_fdir_post_nic_multi_rxq_num} ]; then
	#
	#	log "it rxq number is NOT sw_fdir_post_nic_multi_rxq_num? may it is not UP, please check: cat /proc/interrupts | grep ${ifname}-TxRx | wc -l"
	#
	#	return 0
	#fi
	

	log "phy_configure_fdir_on_sw_post(): begin to adjust irq for ${ifname}"

	cur_irq_cpu_id=0

	for cur_irq in $irq_ids; do

		log "echo ${cur_irq_cpu_id} > /proc/irq/${cur_irq}/smp_affinity_list"

		echo ${cur_irq_cpu_id} > /proc/irq/${cur_irq}/smp_affinity_list 

		cur_irq_cpu_id=`expr ${cur_irq_cpu_id} + ${sw_fdir_post_nic_rxq_cpu_id_step}`

		if [ ${cur_irq_cpu_id} -ge ${cpu_num} ]; then
			cur_irq_cpu_id=0
		fi

	done
}


sw_fdir_post_single_rxq_record_cpu_id=0

	#
	# This "sw_fdir_post_single_rxq_record_cpu_id" is used in post work to interface with single rxq
	#

if [ ${is_configuring_a_single_interface} -eq 1 ]; then

	random_num=`date +%s`

	random_num=`expr ${random_num} % 2`

	
	#
	# If HT on, then sw_fdir_post_single_rxq_record_cpu_id might be:
	#
	#		0, or 4 ( 8 core system ), or 8 ( 16 core system )
	#
	# If HT off, then sw_fdir_post_single_rxq_record_cpu_id should be
	#
	#		always 0
	#

	if [ ${is_running_on_intel_cpu_ht_on} -eq 1 ]; then

		sw_fdir_post_single_rxq_record_cpu_id=`expr ${random_num} \* ${sw_fdir_post_nic_rxq_cpu_id_step}`

	else

		sw_fdir_post_single_rxq_record_cpu_id=0
	fi

fi

log "sw_fdir_post_single_rxq_record_cpu_id=${sw_fdir_post_single_rxq_record_cpu_id}"


phy_configure_fdir_on_sw_post_intel_nic_single_rxq()
{

    ifname=$1
    driver_name=$2

	log "phy_configure_fdir_on_sw_post_intel_nic_single_rxq(): ${ifname}/${driver_name}"

	log "will bind its RXQ to CPU-0 or/and sibiling CPU if CPU HT on"


    cur_irq_cpu_id=$sw_fdir_post_single_rxq_record_cpu_id


	#
	# get IRQ IDs.
	#
	#	intel_nic_single_rxq's IRQ naming pattern is:
	#
	#		eth0-rx-0	
	#		eth0-tx-0
	#

	single_rxq_irq_id=`cat /proc/interrupts | grep ${ifname}-rx-0 | awk '{print $1}' | sed 's/://g'` > /dev/null
	single_txq_irq_id=`cat /proc/interrupts | grep ${ifname}-tx-0 | awk '{print $1}' | sed 's/://g'` > /dev/null


	#
	# set each IRQ's affinity
	#

	log "echo ${cur_irq_cpu_id} > /proc/irq/${single_rxq_irq_id}/smp_affinity_list"

	echo ${cur_irq_cpu_id} > /proc/irq/${single_rxq_irq_id}/smp_affinity_list


	log "echo ${cur_irq_cpu_id} > /proc/irq/${single_txq_irq_id}/smp_affinity_list"


	echo ${cur_irq_cpu_id} > /proc/irq/${single_txq_irq_id}/smp_affinity_list


	#
	# update sw_fdir_post_single_rxq_record_cpu_id
	#

	cur_irq_cpu_id=`expr ${cur_irq_cpu_id} + ${sw_fdir_post_nic_rxq_cpu_id_step}`

	if [ ${cur_irq_cpu_id} -ge ${cpu_num} ]; then
			cur_irq_cpu_id=0
	fi

	sw_fdir_post_single_rxq_record_cpu_id=$cur_irq_cpu_id


	log "now sw_fdir_post_single_rxq_record_cpu_id=${sw_fdir_post_single_rxq_record_cpu_id}"
}


#
# function - phy_configure_fdir_on_sw_post() $1:net_interface $2: driver_name
#
#		check driver name, and call:
#
#			phy_configure_fdir_on_sw_post_intel_nic_multi_rxq()
#		
#			phy_configure_fdir_on_sw_post_intel_nic_single_rxq()
#

phy_configure_fdir_on_sw_post()
{

    ifname=$1
    driver_name=$2

	#
	# step #1. no need to do post
	#

	if [ ${need_sw_fdir_post_on_this_system} -eq 0 ]; then

		log "${gLogPrefix}: NO need to do sw fdir post on this system"

		return 0
	fi


	#
	# step #2. check if interface/driver can and should do post.
	#
	#		and do that post work.
	#

	#
	# step #2.1 --- post work for: intel_nic_multi_rxq
	#

	need_sw_fdir_post_intel_nic_multi_rxq=0

    for post_m_driver in $need_sw_fdir_post_on_this_system_phy_drivers_intel_nic_multi_rxq; do

        if [ "$driver_name" == "$post_m_driver" ]; then
            need_sw_fdir_post_intel_nic_multi_rxq=1
        fi

	done

	if [ $need_sw_fdir_post_intel_nic_multi_rxq -eq 1 ]; then

		phy_configure_fdir_on_sw_post_intel_nic_multi_rxq ${ifname} ${driver_name}


		#
		# Enable SW RSS for in intel_nic_multi_rxq: igb, ixgbe
		#

		log "sw fdir on --- post work: Enable SW RSS for in intel_nic_multi_rxq: ${ifname}/${driver_name}"
	
		if [ -e ${pwd}/configure_if_softrss.sh ]; then
	
			log "sw fdir on --- post work: configure_if_softrss.sh on ${ifname}"
	
			sh ${pwd}/configure_if_softrss.sh on ${ifname}
	
		elif [ -e /bin/appex/if_fdir_rss/configure_if_softrss.sh ]; then
	
			log "sw fdir on --- post work: configure_if_softrss.sh on ${ifname}"
	
			sh /bin/appex/if_fdir_rss/configure_if_softrss.sh on ${ifname}

		else

			log "configure_if_softrss.sh not exist in your system"

		fi

	fi



	#
	# step #2.2 --- post work for: intel_nic_single_rxq
	#
	
	need_sw_fdir_post_intel_nic_single_rxq=0

    for post_m_driver in $need_sw_fdir_post_on_this_system_phy_drivers_intel_nic_single_rxq; do

        if [ "$driver_name" == "$post_m_driver" ]; then
            need_sw_fdir_post_intel_nic_single_rxq=1
        fi

	done

	if [ $need_sw_fdir_post_intel_nic_single_rxq -eq 1 ]; then

		phy_configure_fdir_on_sw_post_intel_nic_single_rxq ${ifname} ${driver_name}
	fi
}


#
# counterpart to phy_configure_fdir_on_sw_post()
#
#
#	For Off, the post work is just rerun RSS script to restore its RXQ and IRQ.
#

phy_configure_fdir_off_sw_post()
{

    ifname=$1
    driver_name=$2


	#
	# step #1. no need to do post
	#

	if [ ${need_sw_fdir_post_on_this_system} -eq 0 ]; then

		log "no need to do sw fdir post on this system"

		return 0
	fi


	#
	# step #2. check if interface/driver can and should do post.
	#
	#		and do that post work.
	#	


	need_sw_fdir_post=0
	need_sw_fdir_post_intel_nic_multi_rxq=0
	need_sw_fdir_post_intel_nic_single_rxq=0


    for post_m_driver in $need_sw_fdir_post_on_this_system_phy_drivers_intel_nic_multi_rxq; do

        if [ "$driver_name" == "$post_m_driver" ]; then
            need_sw_fdir_post=1
			need_sw_fdir_post_intel_nic_multi_rxq=1
        fi

	done


    for post_m_driver in $need_sw_fdir_post_on_this_system_phy_drivers_intel_nic_single_rxq; do

        if [ "$driver_name" == "$post_m_driver" ]; then
            need_sw_fdir_post=1
			need_sw_fdir_post_intel_nic_single_rxq=1
        fi

	done


	if [ $need_sw_fdir_post -eq 1 ]; then


		log "sw fdir off --- post work: atomically run configure_eth_rxq_irq_rss.sh to restore multi-rxq and HW RSS setting for: ${ifname}/${driver_name}"


		pwd=`pwd`

		if [ -e ${pwd}/configure_eth_rxq_irq_rss.sh ]; then

			log "sw fdir off --- post work: configure_eth_rxq_irq_rss.sh ${ifname}"

			sh ${pwd}/configure_eth_rxq_irq_rss.sh ${ifname}

		elif [ -e /bin/appex/if_fdir_rss/configure_eth_rxq_irq_rss.sh ]; then

			log "sw fdir off --- post work: configure_eth_rxq_irq_rss.sh ${ifname}"

			sh /bin/appex/if_fdir_rss/configure_eth_rxq_irq_rss.sh ${ifname}

		else

			log "sw fdir off --- post work: no configure_eth_rxq_irq_rss.sh in your system ?"
		fi


		#
		# Disable SW RSS for intel_nic_multi_rxq, because previously we already restore multi-rxq and HW RSS setting.
		#
		#
		# TODO: this block is only to solve #24865 and #26004. Idealy we should always enable SW RSS
		#		for igb and e1000e on devices whose CPU number is more than NIC maximum rxq,
		#		in order to fully utilize all CPUs.
		#
		#		This is documented in librtc / linuxRtcInterface.c - SoftRss comment.
		#
		#		We may considering do that later, and remove this block.
		#

		if [ $need_sw_fdir_post_intel_nic_multi_rxq -eq 1 ]; then

			log "sw fdir off --- post work: Disable SW RSS for in intel_nic_multi_rxq: ${ifname}/${driver_name}"

			if [ -e ${pwd}/configure_if_softrss.sh ]; then
	
				log "sw fdir off --- post work: configure_if_softrss.sh off ${ifname}"
	
				sh ${pwd}/configure_if_softrss.sh off ${ifname}
	
			elif [ -e /bin/appex/if_fdir_rss/configure_if_softrss.sh ]; then
	
				log "sw fdir off --- post work: configure_if_softrss.sh off ${ifname}"
	
				sh /bin/appex/if_fdir_rss/configure_if_softrss.sh off ${ifname}

			else

				log "configure_if_softrss.sh not exist in your system"

			fi

		fi

	fi
}


#
# function - phy_configure_fdir_on_sw() $1:net_interface $2: driver_name
#
#
#   turn on SW FDir
#
#
#       echo 1 > /proc/net/lw_nic_soft_fdir_${driver_name}/enable
#
#
#       echo '+${ifname}' > /proc/net/lw_nic_soft_fdir_${driver_name}/enable_if
#
#

phy_configure_fdir_on_sw()
{

    ifname=$1
    driver_name=$2
    dir=$(find /proc/net/ -type d -name "*_nic_soft_fdir" | head -n 1)

    if [ -d $dir ]; then

        log "configuring interface: ${ifname}, SW FDir --- on"

        echo 1 > ${dir}/enable

        echo +${ifname} > ${dir}/enable_if


		#
		# SW FDir --- post work
		#

		phy_configure_fdir_on_sw_post ${ifname} ${driver_name}

    else

        log "BAD - configuring interface: ${ifname} / ${driver_name} does not support SW FDir yet, contact DEV"

    fi
}


#
# function - phy_configure_fdir_off_sw() $1:net_interface $2: driver_name
#
#
#   turn off SW FDir
#
#
#       echo -${ifname} > /proc/net/lw_nic_soft_fdir_${driver_name}/enable_if
#
#

phy_configure_fdir_off_sw()
{

    ifname=$1
    driver_name=$2
    
    dir=$(find /proc/net/ -type d -name "*_nic_soft_fdir_${driver_name}" | head -n 1)

    if [ -d $dir ]; then

        log "configuring interface: ${ifname}, SW FDir --- off"

        echo -${ifname} > ${dir}/enable_if


		#
		# SW FDir --- post work
		#

		phy_configure_fdir_off_sw_post ${ifname} ${driver_name}

    else

        log "BAD - configuring interface: ${ifname} / ( ${driver_name} ) does not support SW FDir yet, contact DEV"

    fi
}




#
# function - phy_configure_fdir_off_hw() $1:net_interface $2: driver_name
#
#
#   turn off HW FDir.
#
#       ethtool --features ${ifname} ntuple off
#
#   is common for i40e and other NIC driver (even they don't support HW FDir)
#       

phy_configure_fdir_off_hw()
{

    ifname=$1
    driver_name=$2


	log "ethtool --features ${ifname} ntuple off"
	ethtool --features ${ifname} ntuple off


	if [ "$driver_name" == "i40e" ]; then

		log "phy_configure_fdir_off_hw(): driver(${driver_name}) support HW FDir"

	else

		log "phy_configure_fdir_off_hw(): driver(${driver_name}) does not support HW FDir, fall back to SW FDir"

		phy_configure_fdir_off_sw ${ifname} ${driver_name}
	fi
}



#
# function - phy_configure_fdir_on_hw_i40e() $1:net_interface $2: driver_name
#
#
# For i40e
#
#       turn on HW FDIR    # turn off first, to clear existing FDir rules
#
#       configure 127 rules to match lwconn_v2_hdr->hash_idx ( value range: 0 ~ 127 ), to direct packe to cur_rxq
#
#
# for other
#
#       turn on SW FDIR
#

i40e_hw_fdir_rule_idx_start=0
i40e_hw_fdir_rule_idx_end=127
i40e_hw_fdir_rule_idx_step=1

i40e_hw_fdir_rule_idx_range=$(seq $i40e_hw_fdir_rule_idx_start  $i40e_hw_fdir_rule_idx_step   $i40e_hw_fdir_rule_idx_end)
i40e_hw_fdir_rule_total_cnt=$(expr $i40e_hw_fdir_rule_idx_end - $i40e_hw_fdir_rule_idx_start + 1)


phy_configure_fdir_on_hw_i40e(){
    local ifname=$1
    local driver_name=$2
    local ntuple_status
    local filterRuleIds
    local totalRuleCnt
    local userDefActionMap=()

    log "phy_configure_fdir_on_hw_i40e() - ${ifname}/${driver_name} - turn HW FDIR on, and set HW FDIR rules"

    # no need to set tuple off/on if ntuple already on
    ntuple_status="$(ethtool --show-features ${ifname} 2>/dev/null|awk '/ntuple-filters/{print $2}')"
    [ "${ntuple_status}" != "on" ] && {
        log "ethtool --features ${ifname} ntuple on"

        ethtool --features ${ifname} ntuple on
    }

    totalRuleCnt=$(ethtool --show-ntuple ${ifname} |awk '/^Total/{print $2}')
    filterRuleIds=$(ethtool --show-ntuple ${ifname} |awk '/^Filter:/{print $2}')

    # delete old rule if the count of total rule not equal
    if [ $totalRuleCnt -ne $i40e_hw_fdir_rule_total_cnt ]; then
        for id in ${filterRuleIds}
        do
            log "ethtool --config-ntuple ${ifname} delete ${id}"
            ethtool --config-ntuple ${ifname} delete ${id}
        done
    else
        for id in ${filterRuleIds}
        do
            tmp_maps=($(ethtool --show-ntuple ${ifname} rule $id  |awk '/User-defined:|Action:/{if ($1 == "User-defined:"){usDef=$2} else if ($1 == "Action:") {print usDef, $NF}}'))
            userDefActionMap[${tmp_maps[0]}]=${tmp_maps[1]}
        done
    fi

    cur_rxq_num=`ethtool --show-channels ${ifname} | grep Combined | sed -n 2p | awk '{print $2}'` > /dev/null 2>&1
    cur_rxq_idx=0

	for rule_idx in $i40e_hw_fdir_rule_idx_range; do

        cur_user_def=$((0x20000 + $rule_idx))
        cur_user_def_str=`printf "0x%X" $cur_user_def`

        if [ ${cur_rxq_idx} -eq ${cur_rxq_num} ]; then
            cur_rxq_idx=0
        fi

        isFilterExist=0
        if [ -n "${userDefActionMap[${cur_user_def}]}" ]; then
            if [ "${userDefActionMap[${cur_user_def}]}" -eq "${cur_rxq_idx}" ]; then
                isFilterExist=1
            fi
        fi

        if [ ${isFilterExist} -eq 1 ]; then
            log "this user defined filter(user-def ${cur_user_def_str} action ${cur_rxq_idx}) already exist. no need to config it again."
        else
            log "ethtool --config-ntuple ${ifname} flow-type udp4 user-def ${cur_user_def_str} action ${cur_rxq_idx}"

            ethtool --config-ntuple ${ifname} flow-type udp4 user-def ${cur_user_def_str} action ${cur_rxq_idx}
        fi

        cur_rxq_idx=$(($cur_rxq_idx + 1))

    done
}

phy_configure_fdir_on_hw(){

    ifname=$1
    driver_name=$2

	if [ "$driver_name" == "i40e" ]; then

		log "phy_configure_fdir_on_hw(): driver(${driver_name}) support HW FDir"

		phy_configure_fdir_on_hw_i40e  ${ifname} ${driver_name}

	else

		log "phy_configure_fdir_on_hw(): driver(${driver_name}) does not support HW FDir, fall back to SW FDir"

		phy_configure_fdir_on_sw ${ifname} ${driver_name}
	fi
}


phy_configure_for_one_interface()
{
    ifname=$1

	driver_name=`ethtool -i $ifname | grep driver | awk '{print $2}'`


	#
	# sanity check when set off FDir
	#

	if [ "$gAction" == "off" ]; then

		#
		# check if this phy interface have upper child in wan-tbl.
		#
		# But if gSetOffPhyFromVirtInf==1, then means the turn off request is from upper child, so we skip this check.
		#

		if [ ${gSetOffPhyFromVirtInf} -ne 1 ]; then
	
			if [ ${gSetOffCheckUpperInWanTbl} -eq 1 ]; then
			
				common_have_upper_child_in_wan_tbl ${ifname}
	
				if [ "$?" == "1" ]; then
		
					log "${ifname} - I have upper interfaces in wan-tbl"

					log "${ifname} - dont disable underlying physical interfaces FDir"
		
					return
				
				fi
	
			fi

		fi

	fi
	


	#
	# need to reset ifname to $1, because some previous function call may change it.
	#

    ifname=$1


	#
	# check if this interface's driver is in phy_drivers, then configure it
	#

    for m_driver in $phy_drivers; do

        if [ "$driver_name" != "$m_driver" ]; then
            continue
        fi
		

        if [ "$gAction" == "on" ]; then

			if [ "$gMode" == "sw" ]; then
	            phy_configure_fdir_on_sw ${ifname} ${driver_name}
			fi

			if [ "$gMode" == "hw" ]; then
	            phy_configure_fdir_on_hw ${ifname} ${driver_name}
			fi
        fi

        if [ "$gAction" == "off" ]; then

			if [ "$gMode" == "sw" ]; then
	            phy_configure_fdir_off_sw ${ifname} ${driver_name}
			fi

			if [ "$gMode" == "hw" ]; then
	            phy_configure_fdir_off_hw ${ifname} ${driver_name}
			fi
        fi

    done
}


is_virtual_inf()
{
	ifname=$1

    is_virt=0

    if [ ! -d /sys/class/net/${ifname}/device ]; then
        is_virt=1
    fi

    return ${is_virt}
}


gListPhyInf=""


virt_find_phy_if()
{
	ifname=$1

	is_virtual_inf ${ifname}

	#
	# it is a physical interface, append to gListPhyInf
	#

	if [ "$?" == "0" ]; then

		log "${ifname} is a phy interface, append it to gListPhyInf"

		gListPhyInf="$gListPhyInf $ifname"

		return
	fi

	#
	# it is a virtual interface, recursively find out.
	#

	cd /sys/class/net/${ifname}

	for file in `ls`; do

		if [ ${file:0:6} == "lower_" ]; then

			lower_inf=${file:6}

			virt_find_phy_if ${lower_inf}
		fi
	done
}


#
# check if a virtual interface has siblings, then construct
# gListVirtInfSibilings
#
# for example:
#
#		eth0.100, eth0.200 are siblings.
#

gListVirtInfSibilings=""

virt_find_siblings()
{
	ifname=$1


	if [ ! -d /sys/class/net/${ifname} ]; then
		return
	fi


	cd /sys/class/net/${ifname}

	for file in `ls`; do

		if [ ${file:0:6} == "lower_" ]; then

			lower_parent=${file}

			break
		fi
	done


	if [ ! -d ${lower_parent} ]; then
		return
	fi


	gListVirtInfSibilings=""


	cd ${lower_parent}

	for file in `ls`; do

		if [ ${file:0:6} == "upper_" ]; then

			sibling_inf=${file:6}

			if [ "${sibling_inf}" != "${ifname}" ]; then

				log "${ifname} find sibling_inf ${sibling_inf}, append it to gListVirtInfSibilings"

				gListVirtInfSibilings="${gListVirtInfSibilings} ${sibling_inf}"

			fi
		fi

	done
}


virt_have_siblings()
{
	ifname=$1

	gListVirtInfSibilings=""

	virt_find_siblings $ifname


	if [ "${gListVirtInfSibilings}" == "" ]; then

		log "${ifname} - I don't have virt siblings"

		return 0

	else

		log "${ifname} - my virt siblings are: ${gListVirtInfSibilings}"

		return 1
	fi
}


#
# check if my siblings are in wan-tbl
#

virt_have_siblings_in_wan_tbl()
{
	ifname=$1


	gListVirtInfSibilings=""

	virt_have_siblings $ifname


	if [ "$?" == "0" ]; then

		log "${ifname} - I don't have any virt siblings"

		return 0
	
	fi


	siblings_in_wan_tbl=0

	for sibling_virt_if in ${gListVirtInfSibilings}; do

		log "check if my sibling_virt_if ${sibling_virt_if} in wan-tbl"

		if [ -e /proc/net/lw-conf/wan-tbl ]; then

			cat /proc/net/lw-conf/wan-tbl | grep "${sibling_virt_if}$"

			if [ "$?" == "0" ]; then

				log "my sibling_virt_if ${sibling_virt_if} is in wan-tbl"

				siblings_in_wan_tbl=1
			fi

		fi

	done


	return ${siblings_in_wan_tbl}
}


#
# check if my lower parent is in wan-tbl
#

virt_have_lower_parent_in_wan_tbl()
{
	ifname=$1


	if [ ! -d /sys/class/net/${ifname} ]; then
		return 0
	fi


	lower_parent_inf=""

	cd /sys/class/net/${ifname}

	for file in `ls`; do

		if [ ${file:0:6} == "lower_" ]; then

			lower_parent_inf=${file:6}

			break
		fi
	done


	if [ "${lower_parent_inf}" == "" ]; then
		return 0
	fi 


	if [ -e /proc/net/lw-conf/wan-tbl ]; then

		cat /proc/net/lw-conf/wan-tbl | grep "${lower_parent_inf}$"

		if [ "$?" == "0" ]; then

			log "${ifname} - my parent ${lower_parent_inf} is in wan-tbl"

			return 1
		fi

	fi


	#
	# recurisvely check lower
	#

	virt_have_lower_parent_in_wan_tbl ${lower_parent_inf}

	if [ "$?" == "1" ]; then
		return 1
	fi


	return 0
}


#
# check if my upper child is in wan-tbl
#
# I might be a virtual interface or physical interface
#

common_have_upper_child_in_wan_tbl()
{
	ifname=$1


	if [ ! -d /sys/class/net/${ifname} ]; then
		return 0
	fi


	cd /sys/class/net/${ifname}


	for file in `ls`; do

		if [ ${file:0:6} == "upper_" ]; then

			upper_child=${file:6}

			if [ -e /proc/net/lw-conf/wan-tbl ]; then
		
				cat /proc/net/lw-conf/wan-tbl | grep "${upper_child}$"
		
				if [ "$?" == "0" ]; then
		
					log "${ifname} - my upper_child ${upper_child} is in wan-tbl"
		
					return 1
				fi
		
			fi


			#
			# recurisvely check upper
			#

			common_have_upper_child_in_wan_tbl ${upper_child}

			if [ "$?" == "1" ]; then
				return 1
			fi

		fi

	done


	return 0
}



virt_configure_for_one_interface()
{
	ifname=$1


	#
	# sanity check when set off FDir
	#

	if [ "$gAction" == "off" ]; then


		#
		# have some virtual siblings in wan-tbl
		#

		if [ ${gSetOffCheckSiblingInWanTbl} -eq 1 ]; then

			virt_have_siblings_in_wan_tbl ${ifname}
	
			if [ "$?" == "1" ]; then
	
				log "${ifname} - I have sibling virtual interfaces in wan-tbl"
	
				log "${ifname} - dont disable underlying physical interfaces FDir"
	
				return
			
			fi

		fi


		#
		# have lower parent in wan-tbl
		#

		if [ ${gSetOffCheckLowerInWanTbl} -eq 1 ]; then


			virt_have_lower_parent_in_wan_tbl ${ifname}
	
	
			if [ "$?" == "1" ]; then

				# reset ifname to be $1
	
				ifname=$1

	
				log "${ifname} - I have lower parent interfaces in wan-tbl"
	
				log "${ifname} - dont disable underlying physical interfaces FDir"
	
				return
			
			fi

		fi


		#
		# have upper child in wan-tbl
		#

		if [ ${gSetOffCheckUpperInWanTbl} -eq 1 ]; then
		
			common_have_upper_child_in_wan_tbl ${ifname}

			if [ "$?" == "1" ]; then
	
				log "${ifname} - I have upper interfaces in wan-tbl"
	
				log "${ifname} - dont disable underlying physical interfaces FDir"
	
				return
			
			fi

		fi

	fi


	#
	# clear it first.
	#

	gListPhyInf=""

	virt_find_phy_if $ifname


	#
	# reset ifname to $1, because virt_find_phy_if() may internally change it.
	#

	ifname=$1


	log "virtual interface ${ifname} - gListPhyInf are: ${gListPhyInf}"


	#
	# ** Need to set this flag. **
	#

	gSetOffPhyFromVirtInf=1


	for phy_if in ${gListPhyInf}; do

		log "virtual interface ${ifname} - configure phy_if ${phy_if}"

		phy_configure_for_one_interface ${phy_if}

	done
}



configure_for_one_interface()
{
	ifname=$1


	#
	# /sys/class/net/$ifname is not a directory, it might be 
	#		
	#		/sys/class/net/bonding_masters
	#

	if [ ! -d /sys/class/net/${ifname} ]; then
		return
	fi


	is_virtual_inf ${ifname}

	if [ "$?" == "1" ]; then

		log "$ifname is a virtual interface - find all underlying physical interfaces, and configure them"

		virt_configure_for_one_interface ${ifname}

	else

		log "$ifname is a physical interface - configure it"

		phy_configure_for_one_interface ${ifname}
	fi
}


configure_for_all_interface()
{
	for cur_interface in $all_interfaces; do

		configure_for_one_interface $cur_interface

	done
}


###################################################################################
#
# Main logic starts here
#
###################################################################################



#########################################
#
# find out all interface and their driver
#
#
# For each interface ( if specifiy 'all' as $1 ) or the specified interface ( its ifname as $1 ):
#
#
#	if it is a physical interface, then:
#
#       check if driver is in "phy_drivers", which announce HW FDir or SW FDir support.
#
#       check if interface name is specified by command line argument, or "all".
#
#       configure its FDir, on | off, hw | sw
#
#
#	if it is a virtual interface, then:
#
#		recursively, find all of its lower physical interfaces,
#
#		configure all lower physical interfaces' FDir
#
#
#########################################

all_virtual_interfaces=`ls /sys/devices/virtual/net/`

all_interfaces=`ls /sys/class/net/`


gAction=$2
gMode=$3


log "======= FDir script: $1 $2 $3"


if [ "$1" != "all" ]; then

	log "configure one interface: $1"

	configure_for_one_interface $1

else

	log "configure all interfaces"

	configure_for_all_interface
fi
