#!/bin/sh

# v0.1 - 2021.02.22 - derived from delegate_interrupts.sh
#
# v0.2 - 2021.04.13 - move FDir part into a separate script  configure_eth_fdir.sh
#
# v0.3 - 2021.06.04 - adapt to both bash and busybox ash on arm/mips cpe
#
#					make it easier to extend fors later config logic for interfaces of single rxq
#					and no RSS.
#
# v0.4 - 2021.06.09 - spread NIC IRQ among different CPUs, for arm CPE 2008
#


#
# This script is to:
#
#       configure net_device's  RXQ, RXQ IRQ affinity, RSS.
#
#       including:
#
#			Intel NIC with multi rxq: igb, ixgbe, i40e
#
#			Intel NIC with single rxq: e1000e
#
#			other embedded NIC: dsa, mvneta, fsl_dpa		# TODO
#
#
# Usage:
#
#       this_script         # without argument, configure all net_device.
#
#       this_script  ethX   # one argument, configure the specified net_device.
#                           #
#
#
# The 1st usage should be called in system init time, scanning all interface and configuring them.
#
#
# The 2nd usage should be called when a interface becomes UP ( by "ifconfig ethX up" ) if not previously not UP.
#
#       The reason is:
#
#           mainly for delegate_irq()
#
#           most driver only register IRQ when become UP. That is to say, if not UP, there is no entry in /proc/interrupts, 
#           So the first usage would not run delegate_irq() for such interface.
#
#           when later the interface becomes up, we should configure it again.
#           


gLogFile="/var/log/messages"
gCurrentTime=`date`
gLogPrefix="${gCurrentTime} <LWLog:Info:LW_FDIR_RSS>"

log()
{
	echo "${gLogPrefix}: $@" >> ${gLogFile}
}


#
# all match drivers.
#

match_drivers="i40e ixgbe igb e1000e fsl_dpa"


#
# Intel drivers support RSS and multi-RXQ
#

match_drivers_intel_nic_multi_rxq="i40e ixgbe igb"


#
# Intel drivers support a single RXQ
#

match_drivers_intel_nic_single_rxq="e1000e"

#
# arm CPE 2008's NIC driver
#
match_drivers_arm_cpe_2008="fsl_dpa"


#
# check if running on busybox ash, probably arm/mips CPE
#

is_running_on_busybox=0
if [ -e /bin/busybox ]; then
	is_running_on_busybox=1
fi


#
# 2nd usage: this_script  ethX
#
#		configuring a single interface.
#

is_configuring_a_single_interface=0
if [ $# -eq 1 ]; then
	is_configuring_a_single_interface=1
fi

##########################################
#
# get cpu_num ( system CPU number )
#
##########################################

cpu_num=`cat /proc/cpuinfo | grep "processor" | wc -l` > /dev/null
if [ ${cpu_num} -eq 1 ]; then
    log "system cpu_num: 1, no need to configure rxq,irq,rss"
    exit 0
fi

log "system cpu_num: ${cpu_num}"



#########################
#
# tell if running on CPE 2008 with NXP/Freescale ARM CPU
#
# we can not tell /proc/cpuinfo, because CPE-2008 and CPE-2003,2005 have same cpuinfo keyword
#
#       CPU implementer : 0x41
#       CPU architecture: 8
#       CPU variant     : 0x0
#       CPU part        : 0xd03
#       CPU revision    : 4
#
# So have to tell if there is "QMan portal" in /proc/interrupts, which is unique for
# CPE-2008
#
#########################

is_running_on_arm_cpe_2008=0

tmpvar=`cat /proc/interrupts | grep "QMan portal" | wc -l` > /dev/null

if [ ${tmpvar} -eq 4 ]; then
    is_running_on_arm_cpe_2008=1
fi

log "is_running_on_arm_cpe_2008=${is_running_on_arm_cpe_2008}"



##########################################
#
# stop irqbalance
#
##########################################

irqbalance_exist=0

if [ ${is_running_on_busybox} -eq 0 ]; then

	irqbalance_exist=`ps aux | grep irqbalance` > /dev/null

	if [ -n "${irqbalance_exist}" ]; then

		log "systemctl stop irqbalance"

	    systemctl stop irqbalance > /dev/null
	fi
fi




#
# function - configure_rxq_intel_nic_multi_rxq() $1:net_interface $2: driver_name
#
#		called in parent function configure_rxq()
#
#
#       find max_rxq_num of this interface
#       find cur_rxq_num of this interface
#
#       if cpu_num < max_rxq_num, then set: conf_rxq_num = cpu_num
#
#       otherwise if cpu_num >= max_rxq_num, then set: conf_rxq_num = max_rxq_num
#
#       if conf_rxq_num is equal to cur_rxq_num, no need to exec ethtool command

configure_rxq_intel_nic_multi_rxq()
{
    ifname=$1
    driver_name=$2

    max_rxq_num=`ethtool --show-channels ${ifname} | grep Combined | sed -n 1p | awk '{print $2}'` > /dev/null 2>&1
    cur_rxq_num=`ethtool --show-channels ${ifname} | grep Combined | sed -n 2p | awk '{print $2}'` > /dev/null 2>&1

    if [ ${cpu_num} -lt ${max_rxq_num} ]; then
        conf_rxq_num=${cpu_num}
    else
        conf_rxq_num=${max_rxq_num}
    fi

    if [ ${conf_rxq_num} -eq ${cur_rxq_num} ]; then
        log "rxq num(${conf_rxq_num}) already setting, no need to config it."
    else
        log "ethtool --set-channels ${ifname} combined ${conf_rxq_num}"
        ethtool --set-channels ${ifname} combined ${conf_rxq_num} > /dev/null 2>&1
    fi


    log "configure_rxq_intel_nic_multi_rxq() - ${ifname}(driver: ${driver_name}), set conf_rxq_num to: ${conf_rxq_num}"

}


#
# function - configure_rxq() $1:net_interface $2: driver_name
#
#	
#		check interface driver, and then call:
#
#			configure_rxq_intel_nic_multi_rxq()
#		 

configure_rxq(){

    ifname=$1
    driver_name=$2


	#
	# case #1: intel_nic_multi_rxq
	#

	is_intel_nic_multi_rxq=0

	for rxq_m_driver in $match_drivers_intel_nic_multi_rxq; do

        if [ "$driver_name" == "$rxq_m_driver" ]; then
            is_intel_nic_multi_rxq=1
        fi

	done

	if [ ${is_intel_nic_multi_rxq} -eq 1 ]; then
		configure_rxq_intel_nic_multi_rxq $ifname $driver_name
	fi


	#
	# If non-Intel NIC may support multi rxq, we need to add more logic here.
	#
	# Currently, there is no such NIC.
	#
}



#
# function - configure_rss_intel_nic_multi_rxq() $1:net_interface $2: driver_name
#
#		called in parent function configure_rss()
#
#
#       turn on RSS
#
#       set RSS key to be special symmetric key ( for i40e, a 52 bytes key, for other, 40 bytes key)
#

configure_rss_intel_nic_multi_rxq(){

    ifname=$1
    driver_name=$2

    log "configure_rss_intel_nic_multi_rxq() - ${ifname}(driver: ${driver_name}) - turn RSS on, set RSS sym key"

    ethtool --features ${ifname} rxhash on

    ethtool --config-nfc ${ifname} rx-flow-hash tcp4  sdfn
    ethtool --config-nfc ${ifname} rx-flow-hash udp4  sdfn
    ethtool --config-nfc ${ifname} rx-flow-hash tcp6  sdfn
    ethtool --config-nfc ${ifname} rx-flow-hash udp6  sdfn


    if [ "$driver_name" == "i40e" ]; then

        ethtool --set-rxfh-indir ${ifname} hkey 6d:5a:6d:5a:6d:5a:6d:5a:6d:5a:6d:5a:6d:5a:6d:5a:6d:5a:6d:5a:6d:5a:6d:5a:6d:5a:6d:5a:6d:5a:6d:5a:6d:5a:6d:5a:6d:5a:6d:5a:6d:5a:6d:5a:6d:5a:6d:5a:6d:5a:6d:5a

    else

        ethtool --set-rxfh-indir ${ifname} hkey 6d:5a:6d:5a:6d:5a:6d:5a:6d:5a:6d:5a:6d:5a:6d:5a:6d:5a:6d:5a:6d:5a:6d:5a:6d:5a:6d:5a:6d:5a:6d:5a:6d:5a:6d:5a:6d:5a:6d:5a

    fi
}


#
# function - configure_rss() $1:net_interface $2: driver_name
#
#	
#		check interface driver, and then call:
#
#			is_intel_nic_multi_rxq()
#

configure_rss(){

    ifname=$1
    driver_name=$2


	#
	# case #1: intel_nic_multi_rxq
	#

	is_intel_nic_multi_rxq=0

	for rss_m_driver in $match_drivers_intel_nic_multi_rxq; do

        if [ "$driver_name" == "$rss_m_driver" ]; then
            is_intel_nic_multi_rxq=1
        fi

	done

	if [ ${is_intel_nic_multi_rxq} -eq 1 ]; then
		configure_rss_intel_nic_multi_rxq $ifname $driver_name
	fi


	#
	# If non-Intel NIC may support multi rxq and RSS, we need to add more logic here.
	#
	# Currently, there is no such NIC.
	#
}



#
# function - delegate_irq() $1:net_interface $2: driver_name( unused )
#
#   This function find out corresponding net device queue interrupt ids for net_interface($1) 
#	and delegate them to round-robin count cpu
#
#
#	It internally calls:
#
#		delegate_irq_intel_nic_multi_rxq() $ifname $driver_name
#		
#		delegate_irq_intel_nic_single_rxq() $ifname $driver_name
#


record_cpu_id=0

	#
	# For some Intel NIC with multi rxq, mainly igb, its rxq number is max 4 ( some are 2 ), 
	# usually less than CPU number.
	#
    # if each time we configure a igb's rxq irq affinity to be starting with CPU 0,
	# then the irq of all igb in system would be on first CPU 0 ~ 3. 
	#
    # So we use this variable to delegate **all** igb to CPU 0 ~ N, in all RR manner.
	#
	#
	# [*][*] This "record_cpu_id" variable is also used to delegate CPU for interface with single rxq.
	#

	#
	# If we are using this script to configure a single interface, here we should give
	# record_cpu_id a random value ( mod cpu_num ), instead of CPU-0.
	#
	# otherwise, if we configure many single interfaces many times. their irq would all drop on CPU-0
	#
	# This method give more chance to make single interfaces spread on different CPU.
	#


if [ ${is_configuring_a_single_interface} -eq 1 ]; then

	random_num=`date +%s`

	record_cpu_id=`expr ${random_num} % ${cpu_num}`
fi

log "init record_cpu_id as: ${record_cpu_id}"



delegate_irq_intel_nic_multi_rxq(){

    ifname=$1
    driver_name=$2	

    log "delegate_irq_intel_nic_multi_rxq() - ${ifname}/${driver_name}"


	#
	# get IRQ IDs.
	#
	#	intel_nic_multi_rxq's IRQ naming pattern is:
	#
	#		eth0-TxRx-N
	#

	irq_ids_num=`cat /proc/interrupts | grep ${ifname}-TxRx | wc -l` > /dev/null

	irq_ids=`cat /proc/interrupts | grep ${ifname}-TxRx | awk '{print $1}' | sed 's/://g'` > /dev/null



	#
	# get CPU ID to start with. ( 0, or record_cpu_id )
	#

    cur_irq_cpu_id=0
	
	need_to_use_record_cpu_id=0

	if [ ${irq_ids_num} -lt ${cpu_num} ]; then

		need_to_use_record_cpu_id=1
	fi

	if [ ${need_to_use_record_cpu_id} -eq 1 ]; then
	
		cur_irq_cpu_id=$record_cpu_id
	fi


	#
	# set each IRQ's affinity
	#

	for cur_irq in $irq_ids; do

		log "echo ${cur_irq_cpu_id} > /proc/irq/${cur_irq}/smp_affinity_list"

		echo ${cur_irq_cpu_id} > /proc/irq/${cur_irq}/smp_affinity_list

		cur_irq_cpu_id=`expr ${cur_irq_cpu_id} + 1`

		if [ ${cur_irq_cpu_id} -ge ${cpu_num} ]; then
			cur_irq_cpu_id=0
		fi

	done


	#
	# update record_cpu_id if necessary.
	#

	if [ ${need_to_use_record_cpu_id} -eq 1 ]; then

		record_cpu_id=$cur_irq_cpu_id

		log "now record_cpu_id = ${record_cpu_id}"
	fi
}



delegate_irq_intel_nic_single_rxq(){

    ifname=$1
    driver_name=$2	

    log "delegate_irq_intel_nic_single_rxq() - ${ifname}/${driver_name}"
    

    cur_irq_cpu_id=$record_cpu_id


	#
	# get IRQ IDs.
	#
	#	intel_nic_single_rxq's IRQ naming pattern is:
	#
	#		eth0-rx-0	
	#		eth0-tx-0
	#

	single_rxq_irq_id=`cat /proc/interrupts | grep ${ifname}-rx-0 | awk '{print $1}' | sed 's/://g'` > /dev/null
	single_txq_irq_id=`cat /proc/interrupts | grep ${ifname}-tx-0 | awk '{print $1}' | sed 's/://g'` > /dev/null


	#
	# set each IRQ's affinity
	#

	log "echo ${cur_irq_cpu_id} > /proc/irq/${single_rxq_irq_id}/smp_affinity_list"

	echo ${cur_irq_cpu_id} > /proc/irq/${single_rxq_irq_id}/smp_affinity_list


	log "echo ${cur_irq_cpu_id} > /proc/irq/${single_txq_irq_id}/smp_affinity_list"

	echo ${cur_irq_cpu_id} > /proc/irq/${single_txq_irq_id}/smp_affinity_list


	#
	# update record_cpu_id
	#

	cur_irq_cpu_id=`expr ${cur_irq_cpu_id} + 1`

	if [ ${cur_irq_cpu_id} -ge ${cpu_num} ]; then
		cur_irq_cpu_id=0
	fi

	record_cpu_id=$cur_irq_cpu_id

	log "now record_cpu_id = ${record_cpu_id}"
}


delegate_irq_arm_cpe_2008(){

    #
    # ifname and driver_name are not used in this function.
    #

    ifname=$1
    driver_name=$2  

    log "delegate_irq_arm_cpe_2008()"

#
# /proc/interrupts is:
#
#       33:      10147          0          0      74960     GICv2 216 Level     QMan portal 3
#
#       34:      10175          0      74593          0     GICv2 218 Level     QMan portal 2
#
#       35:      10149      74936          0          0     GICv2 220 Level     QMan portal 1
#
#       36:      84864          0          0          0     GICv2 222 Level     QMan portal 0
#

    irq_ids_num=`cat /proc/interrupts | grep "QMan portal" | wc -l` > /dev/null

    irq_ids=`cat /proc/interrupts | grep "QMan portal" | awk '{print $1}' | sed 's/://g'` > /dev/null


    if [ ${irq_ids_num} -ne ${cpu_num} ]; then

        log "QMan number in /proc/interrupts should be equal to CPU number, something wrong"

        return 0
    fi


    cur_irq_cpu_id=`expr ${cpu_num} - 1`

    for cur_irq in $irq_ids; do

        log "QMan portal ${cur_irq_cpu_id}: echo ${cur_irq_cpu_id} > /proc/irq/${cur_irq}/smp_affinity_list"

        echo ${cur_irq_cpu_id} > /proc/irq/${cur_irq}/smp_affinity_list

        cur_irq_cpu_id=`expr ${cur_irq_cpu_id} - 1`
            
    done
}


delegate_irq(){

    ifname=$1
    driver_name=$2	

	
	#
	# case #1: intel_nic_multi_rxq
	#

	is_intel_nic_multi_rxq=0

	for irq_m_driver in $match_drivers_intel_nic_multi_rxq; do

        if [ "$driver_name" == "$irq_m_driver" ]; then
            is_intel_nic_multi_rxq=1
        fi

	done


	if [ ${is_intel_nic_multi_rxq} -eq 1 ]; then
		delegate_irq_intel_nic_multi_rxq $ifname $driver_name
	fi



	#
	# case #2: intel_nic_single_rxq
	#

	is_intel_nic_single_rxq=0

	for irq_m_driver in $match_drivers_intel_nic_single_rxq; do

        if [ "$driver_name" == "$irq_m_driver" ]; then
            is_intel_nic_single_rxq=1
        fi

	done


	if [ ${is_intel_nic_single_rxq} -eq 1 ]; then
		delegate_irq_intel_nic_single_rxq $ifname $driver_name
	fi


	is_arm_cpe_2008_nic=0

	for irq_m_driver in $match_drivers_arm_cpe_2008; do

		if [ "$driver_name" == "$irq_m_driver" ]; then
			if [ ${is_running_on_arm_cpe_2008} -eq 1 ]; then
				is_arm_cpe_2008_nic=1
			fi
		fi
	done

	if [ ${is_arm_cpe_2008_nic} -eq 1 ]; then
		delegate_irq_arm_cpe_2008 $ifname $driver_name
	fi


	#
	# TODO: we need to delegate_irq() for
	#
	#			interfaces on arm/mips cpe with single rxq.
	#
	# those interfaces single irq need to be spread on different cores, not default CPU 0.
	#
}


#########################################
#
#
# Main logic:
#
#
# 	for each NIC:
#
#
#       #1. configure_rxq()
#
#
#       #2. delegate_irq()
#
#
#       #3. configure_rss()
#
#
#########################################

all_virtual_interfaces=`ls /sys/devices/virtual/net/`

all_interfaces=`ls /sys/class/net/`


is_virtual_inf()
{
    ifname=$1

    is_virt=0

    if [ ! -d /sys/class/net/${ifname}/device ]; then
        is_virt=1
    fi

    return ${is_virt}
}



for cur_interface in $all_interfaces; do

	#
	# skip any virtual interfaces, only handle physical interfaces
	#

    is_virtual_inf $cur_interface

    if [ "$?" == "1" ]; then

        log "skip virtual interface: ${cur_interface}"

        continue
    fi


	#
	# get interface driver
	#

    cur_driver=`ethtool -i $cur_interface | grep driver | awk '{print $2}'`


	log "checking interface: device: ${cur_interface}, driver: ${cur_driver}"


	#
	# check if this interface's driver is in match_drivers, then configure its RXQ/RSS/IRQ affinity
	#

    for m_driver in $match_drivers; do


        if [ "$cur_driver" != "$m_driver" ]; then
            continue
        fi


		#
		# is_configuring_a_single_interface == TRUE
		#

		if [ $# -eq 1 ]; then

			if [ "$1" != "$cur_interface" ]; then
				continue
			fi
		fi


		log "configuring interface: ${cur_interface}, driver: ${cur_driver}"

		
		configure_rxq ${cur_interface} ${cur_driver}

	    delegate_irq ${cur_interface} ${cur_driver}

    	configure_rss ${cur_interface} ${cur_driver}


    done

done
