#!/bin/sh

#
# #25391 - separately upload shaping SA msg in NIC driver. - called "sepsa" for short
#
#	this script is to configure behavior of NIC driver.
#
#
#	each NIC driver support sepsa would have this procfs directory
#
#		/proc/net/lw_nic_${drivername}		# like lw_nic_i40e
#
#			sepsa-enable-if		# R/W: echo "+|- ifname" to enable/disable sepsa on this interface
#
#			sepsa-master-ip		# R/W: echo SAMasterIp into it
#
#			sepsa-master-port	# R/W: echo SAMasterPort into it
#
#			sepsa-stats			# R/W: write to clear
#
#
#	This script is to configure NIC driver by these procfs interface.
#
#
#	2 Usage of this script:
#
#		#1. this_script  on  ifname  SAMasterIp  SAMasterPort
#
#			this usage is called when CPE become a master.
#
#			ifname can be a phy_if or virtual_if
#
#			if a virtual_if, then all underlying phy_if are enabled with sepsa.
#
#
#		#2. this_script off
#
#			disable all phy_if's sepsa.
#
#			this usage is called when CPE become a slave.
#
#			also called when agent startup, because if agent crash/reboot, remaining sepsa setting
#			on phy_if need to be cleaned.
#
#
# ----------------------------
#
# history:
#
#	v0.1 - 2021.07.23
#



gLogFile="/var/log/messages"
gCurrentTime=`date`
gLogPrefix="${gCurrentTime} <LWLog:Info:LW_SEPSA>"

log()
{
	echo "${gLogPrefix} $@" >> ${gLogFile}
}


#
# their initial value are invalid, should be overwriten by cmd-line arguments later.
#

gSAMasterIp=256.256.256.256
gSAMasterPort=63356


###########################################
#
# variable declare.
#
###########################################

#
# drivers supports sepsa
#

phy_drivers="i40e ixgbe igb"



##########################################
#
# functions
#
##########################################


#
# function - phyif_set_on() $1:ifname
#
#
#   turn on sepsa of specific physical_if
#
#		echo gSAMasterIp > /proc/net/lw_nic_${drvname}/sepsa-master-ip
#
#		echo gSAMasterPort > /proc/net/lw_nic_${drvname}/sepsa-master-port
#
#       echo +${ifname} > /proc/net/lw_nic_${drvname}/sepsa-enable-if
#

phyif_set_on()
{
    ifname=$1

	drvname=`ethtool -i $ifname | grep driver | awk '{print $2}'`

	log "phyif_set_on(): ${ifname}"


	if [ "$drvname" == "" ]; then

		log "can not get drvname of ${ifname}, skip"

		return
	fi

    dir=$(find /proc/net/ -type d -name "*_nic_${drvname}" | head -n 1)

	if [ ! -e "${dir}/sepsa-enable-if" ]; then

		log "${ifname}/${drvname} does not support sepsa, skip"

		return
	fi


	echo "${gSAMasterIp}" > ${dir}/sepsa-master-ip

	if [ "$?" != "0" ]; then

		log "SAMasterIp: ${gSAMasterIp} is invalid"

		return
	fi

	
	echo "${gSAMasterPort}" > ${dir}/sepsa-master-port

	if [ "$?" != "0" ]; then

		log "gSAMasterPort: ${gSAMasterPort} is invalid"

		return
	fi


	echo "+${ifname}" > ${dir}/sepsa-enable-if

	if [ "$?" != "0" ]; then

		log "${ifname} is not owned by this driver ${drvname}"

		return
	fi


	log "phyif_set_on(): ${ifname}/${drvname}, done"
}


#
# function - phyif_set_off() $1:ifname
#
#
#   turn off sepsa of specific physical_if
#
#       echo -${ifname} > /proc/net/lw_nic_${drvname}/sepsa-enable-if
#

phyif_set_off()
{
    ifname=$1

	drvname=`ethtool -i $ifname | grep driver | awk '{print $2}'`


	log "phyif_set_off(): ${ifname}/${drvname}"


	if [ "$drvname" == "" ]; then

		log "can not get drvname of ${ifname}, skip"

		return
	fi

    dir=$(find /proc/net/ -type d -name "*_nic_${drvname}" | head -n 1)

	if [ ! -e "${dir}/sepsa-enable-if" ]; then

		log "${ifname}/${drvname} does not support sepsa, skip"

		return
	fi


	echo "-${ifname}" > ${dir}/sepsa-enable-if

	if [ "$?" != "0" ]; then

		log "${ifname} is not owned by this driver ${drvname}"

		return
	fi


	log "phyif_set_off(): ${ifname}/${drvname}, done"	
}



all_virtual_interfaces=`ls /sys/devices/virtual/net/`

all_interfaces=`ls /sys/class/net/`


#
# function - is_virtif() $1:ifname
#
#	tell if a interface is a virtual interface.
#

is_virtif()
{
	ifname=$1

    is_virt=0

	if [ ! -d /sys/class/net/${ifname}/device ]; then
		is_virt=1
	fi

    return ${is_virt}
}


gListPhyInf=""


#
# function - virtif_find_phyif() $1:ifname
#
#	find all underlying physical interfaces of a virtual interface.
#
#	construct them in gListPhyInf
#

virtif_find_phyif()
{
	ifname=$1

	log "virtif_find_phyif(): ${ifname}"
	

	is_virtif ${ifname}

	#
	# it is a physical interface, append to gListPhyInf
	#

	if [ "$?" == "0" ]; then

		log "${ifname} is a phy interface, append it to gListPhyInf"

		gListPhyInf="$gListPhyInf $ifname"

		return
	fi

	#
	# it is a virtual interface, recursively find out.
	#

	cd /sys/class/net/${ifname}

	for file in `ls`; do

		if [ ${file:0:6} == "lower_" ]; then

			lower_inf=${file:6}

			virtif_find_phyif ${lower_inf}
		fi

	done
}


#
# function - phyif_set_off_all_interfaces()
#
#   turn off sepsa of all physical_if
#

phyif_set_off_all_interfaces()
{

	log "phyif_set_off_all_interfaces()"


	for cur_interface in $all_interfaces; do

		#
		# skip virtual interfce
		#

		is_virtif ${cur_interface}
	
		if [ "$?" == "1" ]; then

			continue;
		fi


		phyif_set_off ${cur_interface}
		
	done


	log "phyif_set_off_all_interfaces(), done"
}


#
# function - common_set_on_one_interface()
#
#	turn on sepsa on one interface, it can be phy_if or virt_if
#

common_set_on_one_interface()
{
	cur_interface=$1

	log "common_set_on_one_interface(): ${cur_interface}"


	is_virtif ${cur_interface}
	

	if [ "$?" == "0" ]; then

	#
	# phy_if, set it on
	# 

		log "${cur_interface} is a phy_if"

		phyif_set_on ${cur_interface}

	else

	#
	# virt_if, set on all underlying phy_if
	#

		log "${cur_interface} is a virt_if"


		gListPhyInf=""
	
		virtif_find_phyif $ifname		

		log "gListPhyInf: ${gListPhyInf}"


		for phy_if in ${gListPhyInf}; do
		
			phyif_set_on ${phy_if}
	
		done

	fi


	log "common_set_on_one_interface(): ${cur_interface}, done"
}


##########################################
#
# main logic:
#
#	this_script  on  ifname  SAMasterIp  SAMasterPort
#
#		turn on sepsa of one interface ( if a virt_if, then all underlying phy_if ).
#
#
#	this_script off
#
#		turn off sepsa of all phy_if
#		
#
##########################################

# check arguments.

if [ $# -lt 1 ]; then
    log "bad script usage"
    exit 1
fi


#
#	this_script off
#
#		turn off sepsa of all phy_if
#

if [ "$1" == "off" ]; then
	
	phyif_set_off_all_interfaces

	exit 0
fi


#
#	this_script  on  ifname  SAMasterIp  SAMasterPort
#
#		turn on sepsa of one interface ( if a virt_if, then all underlying phy_if ).
#

if [ "$1" != "on" ]; then
    log "bad script usage: 1st argument must be on/off"
	exit 1
fi


if [ $# -ne 4 ]; then
    log "bad script usage: if set on, must input 4 arguments"
    exit 1
fi


#
# these 2 are used internally in set on function.
#

gSAMasterIp=$3
gSAMasterPort=$4

common_set_on_one_interface $2
