#!/bin/sh

#
# #25624 - Soft RSS
#
#	this script is to configure SoftRSS of NIC driver.
#
#
#	each NIC driver support SoftRSS would have this procfs directory
#
#		/proc/net/lw_nic_${drivername}		# like lw_nic_i40e
#
#			softrss-enable-if	# R/W: echo "+|- ifname" to enable/disable soft RSS on this interface
#
#			softrss-stats			# R/W: write to clear
#
#
#	This script is to configure NIC driver by these procfs interface.
#
#
#	2 Usage of this script:
#
#		#1. this_script  on  ifname
#
#			ifname can be a phy_if or virtual_if
#
#			if a virtual_if, then all underlying phy_if are enabled with soft RSS.
#
#
#		#2. this_script off  ifname | all
#
#			disable all phy_if's softrss
#
#			also called when agent startup, because if agent crash/reboot, remaining softrss setting
#			on phy_if need to be cleaned.
#
#
# ----------------------------
#
# history:
#
#	v0.1 - 2021.07.23
#



gLogFile="/var/log/messages"
gCurrentTime=`date`
gLogPrefix="${gCurrentTime} <LWLog:Info:LW_SOFTRSS>"

log()
{
	echo "${gLogPrefix} $@" >> ${gLogFile}
}


###########################################
#
# variable declare.
#
###########################################

#
# drivers supports softrss.
#
# TODO: add e1000e later.
#

phy_drivers="igb ixgbe"



##########################################
#
# functions
#
##########################################


#
# function - phyif_set_on() $1:ifname
#
#
#   turn on softrss of specific physical_if
#
#       echo +${ifname} > /proc/net/lw_nic_${drvname}/softrss-enable-if
#

phyif_set_on()
{
    ifname=$1

	drvname=`ethtool -i $ifname | grep driver | awk '{print $2}'`


	log "phyif_set_on(): ${ifname}"


	if [ "$drvname" == "" ]; then

		log "can not get drvname of ${ifname}, skip"

		return
	fi

    dir=$(find /proc/net/ -type d -name "*_nic_${drvname}" | head -n 1)

	if [ ! -e "${dir}/softrss-enable-if" ]; then

		log "${ifname}/${drvname} does not support softrss, skip"

		return
	fi


	echo "+${ifname}" > ${dir}/softrss-enable-if

	if [ "$?" != "0" ]; then

		log "${ifname} is not owned by this driver ${drvname}"

		return
	fi


	log "phyif_set_on(): ${ifname}/${drvname}, done"
}


#
# function - phyif_set_off() $1:ifname
#
#
#   turn off softrss of specific physical_if
#
#       echo -${ifname} > /proc/net/lw_nic_${drvname}/softrss-enable-if
#

phyif_set_off()
{
    ifname=$1

	drvname=`ethtool -i $ifname | grep driver | awk '{print $2}'`


	log "phyif_set_off(): ${ifname}/${drvname}"


	if [ "$drvname" == "" ]; then

		log "can not get drvname of ${ifname}, skip"

		return
	fi

    dir=$(find /proc/net/ -type d -name "*_nic_${drvname}" | head -n 1)

	if [ ! -e "${dir}/softrss-enable-if" ]; then

		log "${ifname}/${drvname} does not support softrss, skip"

		return
	fi


	echo "-${ifname}" > ${dir}/softrss-enable-if

	if [ "$?" != "0" ]; then

		log "${ifname} is not owned by this driver ${drvname}"

		return
	fi


	log "phyif_set_off(): ${ifname}/${drvname}, done"	
}



all_virtual_interfaces=`ls /sys/devices/virtual/net/`

all_interfaces=`ls /sys/class/net/`


#
# function - is_virtif() $1:ifname
#
#	tell if a interface is a virtual interface.
#

is_virtif()
{
	ifname=$1

    is_virt=0

	if [ ! -d /sys/class/net/${ifname} ]; then
		is_virt=0
		return ${is_virt}
	fi


	if [ ! -d /sys/class/net/${ifname}/device ]; then
		is_virt=1
		return ${is_virt}
	fi

    return ${is_virt}
}


gListPhyInf=""


#
# function - virtif_find_phyif() $1:ifname
#
#	find all underlying physical interfaces of a virtual interface.
#
#	construct them in gListPhyInf
#

virtif_find_phyif()
{
	ifname=$1

	log "virtif_find_phyif(): ${ifname}"
	

	is_virtif ${ifname}

	#
	# it is a physical interface, append to gListPhyInf
	#

	if [ "$?" == "0" ]; then

		log "${ifname} is a phy interface, append it to gListPhyInf"

		gListPhyInf="$gListPhyInf $ifname"

		return
	fi

	#
	# it is a virtual interface, recursively find out.
	#

	cd /sys/class/net/${ifname}

	for file in `ls`; do

		if [ ${file:0:6} == "lower_" ]; then

			lower_inf=${file:6}

			virtif_find_phyif ${lower_inf}
		fi

	done
}


#
# function - phyif_set_off_all_interfaces()
#
#   turn off softrss of all physical_if
#

phyif_set_off_all_interfaces()
{

	log "phyif_set_off_all_interfaces()"


	for cur_interface in $all_interfaces; do

		#
		# skip virtual interfce
		#

		is_virtif ${cur_interface}
	
		if [ "$?" == "1" ]; then

			continue;
		fi


		phyif_set_off ${cur_interface}
		
	done


	log "phyif_set_off_all_interfaces(), done"
}


#
# function - common_set_on_one_interface()
#
#	turn on softrss on one interface, it can be phy_if or virt_if
#

common_set_on_one_interface()
{
	cur_interface=$1

	log "common_set_on_one_interface(): ${cur_interface}"


	is_virtif ${cur_interface}
	

	if [ "$?" == "0" ]; then

	#
	# phy_if, set it on
	# 

		log "${cur_interface} is a phy_if"

		phyif_set_on ${cur_interface}

	else

	#
	# virt_if, set on all underlying phy_if
	#

		log "${cur_interface} is a virt_if"


		gListPhyInf=""
	
		virtif_find_phyif $ifname		

		log "gListPhyInf: ${gListPhyInf}"


		for phy_if in ${gListPhyInf}; do
		
			phyif_set_on ${phy_if}
	
		done

	fi


	log "common_set_on_one_interface(): ${cur_interface}, done"
}


#
# function - common_set_off_one_interface()
#
#	turn off softrss on one interface, it can be phy_if or virt_if
#

common_set_off_one_interface()
{
	cur_interface=$1

	log "common_set_off_one_interface(): ${cur_interface}"


	is_virtif ${cur_interface}
	

	if [ "$?" == "0" ]; then

	#
	# phy_if, set it off
	# 

		log "${cur_interface} is a phy_if"

		phyif_set_off ${cur_interface}

	else

	#
	# virt_if, set on all underlying phy_if
	#

		log "${cur_interface} is a virt_if"


		gListPhyInf=""
	
		virtif_find_phyif $ifname		

		log "gListPhyInf: ${gListPhyInf}"


		for phy_if in ${gListPhyInf}; do
		
			phyif_set_off ${phy_if}
	
		done

	fi


	log "common_set_off_one_interface(): ${cur_interface}, done"
}


##########################################
#
# main logic:
#
#	this_script  on  ifname
#
#		turn on softrss of one interface ( if a virt_if, then all underlying phy_if ).
#
#
#	this_script off ifname | all
#
#		turn off softrss of all phy_if or specified interfaces ( can be virt_if )
#		
#
##########################################

#
# get cpu_num ( system CPU number )
#

cpu_num=`cat /proc/cpuinfo | grep "processor" | wc -l` > /dev/null
if [ ${cpu_num} -eq 1 ]; then
    log "system cpu_num: 1, no need to configure soft RSS"
    exit 0
fi


log "system cpu_num: ${cpu_num}"


#
# check arguments.
#

if [ $# -ne 2 ]; then
    log "bad script usage"
    exit 1
fi


#
# this_script  on  ifname
#

if [ "$1" == "on" ]; then

	common_set_on_one_interface $2

	exit 0
fi


if [ "$1" == "off" ]; then

	if [ "$2" == "all" ]; then

		phyif_set_off_all_interfaces

	else

		common_set_off_one_interface $2
	fi

fi
